/*
 * ssse-sample.cpp
 *
 * Copyright (c) 2016 Ryu Takeda
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/mit-license.php
 *
 */

#include <iostream>
#include <random>
#include <sys/time.h>
#include <tmmintrin.h>

double get_time() {
    struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec + tv.tv_usec * 1.0E-6;
}

template<typename T>
inline signed char float2char(T x, T scale) {
	return (signed char) (127.0 * x / scale + 0.5);
}

template<typename T>
inline unsigned char sigmoid2uchar(T x) {
	return (unsigned char) (255.0 * x + 0.5);
}

/**
 *  @param qmat row-major
 *  @param fmat row-major
 */
template<typename T>
void mat2char(signed char *qmat, T *scale, //
			  T *fmat, int n_row, int n_col) {
	for (int i = 0; i < n_row; i++) {
		T s_max = -1.0;
		for (int j = 0; j < n_col; j++) {
			if (s_max < fabs(fmat[n_col * i + j])) {
				s_max = fabs(fmat[n_col * i + j]);
			}
		}
		scale[i] = s_max;
		for (int j = 0; j < n_col; j++) {
			qmat[n_col * i + j] = float2char(fmat[n_col * i + j], s_max);
		}
	}
}

template<typename T>
void float_matXvec(T *out, //
				   T *mat, T *vec, int n_row, int n_col) {	
	for (int i = 0; i < n_row; i++) {
		T sum = 0;
		for (int j = 0; j < n_col; j++) {
			sum += mat[n_col * i + j] * vec[j];
		}
		out[i] = sum;
	}
}

void ssse_matXvec(int *out, //
				  signed char *qmat, unsigned char *qvec, int n_row, int n_col) {
	for (int i = 0; i < n_row; i++) {
		int sum = 0; 

		// here, hard-coded.
		// the number of column should be k * 16 (k = 1, 2, ...).
		for (int j = 0; j < n_col / 16; j++) { 
			int result[4] = { 0, 0, 0, 0 };
			__m128i m_qin, m_qw;
			m_qin = _mm_load_si128((__m128i *) &qvec[j * 16]);
			m_qw = _mm_load_si128((__m128i *) &qmat[n_col * i + j * 16]);			
			__m128i sum8 = _mm_set_epi32(0, 0, 0, 0);
			{
				__m128i c = _mm_maddubs_epi16(m_qin, m_qw);
				__m128i lo = _mm_srai_epi32(_mm_unpacklo_epi16(c, c), 16);
				__m128i hi = _mm_srai_epi32(_mm_unpackhi_epi16(c, c), 16);				
				sum8 = _mm_add_epi32(_mm_add_epi32(lo, hi), sum8);
			}
			_mm_store_si128((__m128i *) result, sum8);
			sum += result[0] + result[1] + result[2] + result[3];
		}
		out[i] = sum;		
	}
}


typedef float data_t;

int main(int argc, char *argv[])
{
	int n_row = 4096, n_col = 4096, *ires;
	data_t *fmat, *fvec, *fres, *scale, *qres;
	signed char *qmat;
	unsigned char *qvec;

	if (argc < 2) {
		printf("usage: matsize\n");
		return -1;
	}

	n_row = n_col = atoi(argv[1]);
	 
    // memory allocation
	{
		fmat = (data_t *)calloc(n_row * n_col, sizeof(data_t));
		fvec = (data_t *)calloc(n_col, sizeof(data_t));
		fres = (data_t *)calloc(n_row, sizeof(data_t));

		posix_memalign((void **) &qmat, 16,
					   sizeof(signed char) * n_row * n_col);
		posix_memalign((void **) &qvec, 16, 
					   sizeof(unsigned char) * n_col);
		scale = (data_t *)calloc(n_row, sizeof(data_t));
		qres = (data_t *)calloc(n_row, sizeof(data_t));
		ires = (int *)calloc(n_row, sizeof(int));		
	}
	
	std::mt19937 mt(777);
	std::uniform_real_distribution<data_t> genuni(0.0, 1.0);
	std::normal_distribution<data_t> gennorm(0.0, 2.0);

	{
		// data generation
		for (int i = 0; i < n_row * n_col; i++) {
			fmat[i] = gennorm(mt);
		}
		for (int i = 0; i < n_col; i++) {
			fvec[i] = genuni(mt);
		}
		mat2char(qmat, scale, fmat, n_row, n_col);

        // floating point
		double f_start = get_time();
		{
			float_matXvec(fres, fmat, fvec, n_row, n_col);
		}
		double f_end = get_time();

		// fixed point
		double q_start = get_time();
		{
			for (int i = 0; i < n_row; i++) {
				qvec[i] = sigmoid2uchar(fvec[i]);
			}			
			ssse_matXvec(ires, qmat, qvec, n_row, n_col);
			for (int i = 0; i < n_row; i++) {
				qres[i] = ires[i] * scale[i] / 255.0 / 127.0;
			}
		}
		double q_end = get_time();

		printf("proc-time: float %.3E sec, fixed: %.3E sec\n", 
			   f_end - f_start, q_end - q_start);
	}

    // release
	free(fmat);
	free(fvec);
	free(fres);
	free(qmat);
	free(qvec);
    free(scale);
	free(ires);
    free(qres);
	return 0;
}

